/*
 * Copyright (c) 2009-2010 Jose-Luis Lisani <joseluis.lisani@uib.es>
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file colorfilteringIPOL.cpp
 *
 * @brief Various functions for RGB cube processing 
 *        (including filtering and density computation)
 *
 * @version 1.0 
 * @author Jose Luis Lisani; <joseluis.lisani@uib.es> 
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "mdmath.h"
#include "miscIPOL.h"
#include "colorfilteringIPOL.h"


/**
* \brief Data structure containing RGB information for each image pixel 
*
* \param R, G, B        color point associated to the pixel (RGB values at the 
*                       pixel)
* \param AA, BB, CC, DD parameters of a local plane around the color point in 
*                       RGB space
* \param Rcm, Gcm, Bcm  center of mass of the color points in the neighborhood 
*                       of the considered color point 
* \param processed      label indicating whether the color point has been
*                       already processed (processed=1) or not (processed=0) 
* \param projected      label indicating whether the color point can be
*                       projected (projected=1) or not (projected=0) onto a
*                       line or plane 
* \param ignore         label indicating whether the color point may be
*                       ignored (ignore=1) or not (ignore=0) for processing 
*
*/
struct infoPixel {
  float R, G, B; //color point associated to the pixel 
                 //(RGB values at the pixel) 
  //info local plane
  float AA, BB, CC, DD; //parameters of a local plane around the color point in
                        // RGB space
  float Rcm, Gcm, Bcm; //center of mass of the color points 
                       //in the neighborhood 
                       //of the considered color point 
  unsigned char processed; //label indicated whether the color point has been 
                           //already processed (processed=1) 
                           //or not (processed=0) 
  unsigned char projected; //label indicated whether the color point can be 
                           //projected (projected=1) or not (projected=0) onto
                           //a line or plane 
  unsigned char ignore;  //label indicated whether the color point may be
                         //ignored (ignore=1) or not (ignore=0) for processing 
};


/**
* \brief Data structure containg information of a color point in RGB space
*
* \param R, G, B           RGB values of the point 
* \param idpix             identifier of the image pixel with color RGB 
* \param next, nextneigh   pointers to other data structures of the same type
*
*/
struct RGBpoint {
  float R, G, B; //RGB values of the point 
  int idpix; //identifier of the image pixel with color RGB 
  struct RGBpoint *next, *nextneigh; //pointers to other data structures of 
                                     //the same type
};

/**
* \brief List of data structures of type RGBpoint
*
* \param first, last    pointers to first and last elements of the list 
*/
struct RGBbin {
  struct RGBpoint *first, *last; //pointers to first and last elements of the 
                                 //list
};

/** 
* \brief Information about the projection of an RGB color point 
*        onto a line or plane
*
* \param  Rp, Gp, Bp          coordinates of the projected point  
* \param  A, B, C, D          parameters of the projection line or plane  
* \param  weight1, weight2    the two largest eigenvalues from Principal 
*                             Components Analysis of the distribution of 
*                             colors in the neighborhood of the considered 
*                             RGB point   
* \param  Rcm, Gcm, Bcm       center of mass of the colors in the neighborhood 
*                             of the RGB color point   
*
*/
struct infoProjection {
  float Rp, Gp, Bp; //coordinates of the projected point 
  float A, B, C, D; //parameters of the projection line or plane 
  float weight1, weight2; //the two largest eigenvalues from Principal 
                          //Components Analysis of the distribution of 
                          //colors in the neighborhood of the considered 
                          //RGB point 
  float Rcm, Gcm, Bcm; //center of mass of the colors in the neighborhood 
                       //of the RGB color point   
};

/**
* \brief initialization of data structures for RGB cube processing
*
* @param[in]  R, G, B   pointers to red, green and blue channels of the input 
*                       image
* @param[out] checkOut  checkOut[iR][iG][iB] is a pointer to a RGBbin data 
*                       structure that contains the list of pixels whose color
*                       is in the range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[out] infopixOut infopixOut[k] is a pointer to a infoPixel data 
*                        structure that contains information about the color 
*                        of pixel k
* @param[out] infopixOutB infopixOuBt[k] is a pointer to a infoPixel data 
*                         structure that contains information about the color 
*                         of pixel k
* @param[out] repeatedRGBOut repeatedRGBOut[k] counts the number of
*                            occurrences of the color of pxel k, provided
*                            that pixel k is the first occurrence of the
*                            color, else repeatedRGBOut[k]=0 
* @param[in] ignore000 	if ignore000=1 do not take into account pixels with 
*                       (0, 0, 0) color (eventually excluded from processing 
*                       after a preprocessing step)
*
* Remark:
* Initially infopixOut and infopixOutB contain the same information, 
* but as the algorithm iterates one of them is used to store the update
* information while the other keeps the old RGB values. Their role switches 
* as the algorithm iterates.
*
* \return number of colors in the image 
*         (always number of colors <= number of pixels)
*/
int initialize_colorfiltering_data(unsigned char *R, unsigned char *G, 
                                   unsigned char *B, int w, int h, 
                                   struct RGBbin ****checkOut, 
                                   struct infoPixel **infopixOut, 
                                   struct infoPixel **infopixBOut, 
                                   int **repeatedRGBOut, 
                                   unsigned char ignore000)
{									
  struct RGBbin ***check;
  struct infoPixel *infopix, *infopixB=NULL;
  struct RGBpoint *point;
  int *repeatedRGB;
  int iR, iG, n, nRGB;

  //create pointers to RGBbin data structures 
  check=new struct RGBbin**[256];
  for (iR=0; iR < 256; iR++) {
    check[iR]=new struct RGBbin*[256];
    for (iG=0; iG < 256; iG++) {
      check[iR][iG]=new struct RGBbin[256];
      memset(check[iR][iG], 0, 256*sizeof(struct RGBbin));
    }
  }  

  //fill information about the colors of the pixels
  nRGB=0;
  infopix=new struct infoPixel[w*h];
  if (infopixBOut) infopixB=new struct infoPixel[w*h];
  unsigned char isGray=1;
  for (n=0; n < w*h; n++) {
    infopix[n].R=(float) R[n];
    infopix[n].G=(float) G[n];
    infopix[n].B=(float) B[n];
    //check gray value
    isGray= (isGray && ((R[n] == G[n]) && (R[n] == B[n])));
    infopix[n].processed=0;
    infopix[n].projected=0;
    //if option ignore000 is active (ignore000=1) do not take into account 
    //(0, 0, 0) pixels (usually belonging to a mask)
    infopix[n].ignore=0;
    if (ignore000 && ((R[n] == 0) && (G[n] == 0) && (B[n] == 0))) { 
      infopix[n].ignore=1;
    } else {
      //create or update list of color points
      point=new struct RGBpoint;
      point->R=(float) R[n];
      point->G=(float) G[n];
      point->B=(float) B[n];
      point->idpix=n; //identifier of the pixel where the color is found
      point->next=NULL;
      if (check[R[n]][G[n]][B[n]].first == NULL) nRGB++;
      if (check[R[n]][G[n]][B[n]].first == NULL) 
        check[R[n]][G[n]][B[n]].first=point;
      else check[R[n]][G[n]][B[n]].last->next=point;
      check[R[n]][G[n]][B[n]].last=point;
    }
  }

  //if it is gray image, then do not filter
  if (isGray) {
    for (n=0; n < w*h; n++) infopix[n].ignore=1;
    printf("Gray level image!\n");
    printf("\nNumber of gray levels of input image: %i\n", nRGB);
    printf("Do not filter\n");
  } else {
    printf("\nNumber of colors of input image: %i\n", nRGB);
  }

  //return values
  *checkOut=check;
  *infopixOut=infopix;
  if (infopixBOut) *infopixBOut=infopixB;

  if (repeatedRGBOut) {
    //count voxels
    //repeatedRGB[n]=number of times that the value of pixel n 
    //is repeated, if n is the first instance of the value
    //else repeatedRGB[n]=0
    repeatedRGB=new int[w*h];
    memset(repeatedRGB, 0, w*h*sizeof(int));
    for (int iR=0; iR < 256; iR++)
      for (int iG=0; iG < 256; iG++)
        for (int iB=0; iB < 256; iB++) {
          if (check[iR][iG][iB].first) { 
            repeatedRGB[check[iR][iG][iB].first->idpix]=1; //first is not 
                                                           //repeated
            for (point=check[iR][iG][iB].first->next; point; 
                 point=point->next) {
              repeatedRGB[point->idpix]=0;
              repeatedRGB[check[iR][iG][iB].first->idpix]++; //count number of 
                                                             //repetitions
            }
          }
        }
    *repeatedRGBOut=repeatedRGB;
  }

  //number of different colors
  return nRGB;

}


/**
* \brief free memory used to store temporary data structures
*/
void delete_colorfiltering_data(struct RGBbin ***check, 
                                struct infoPixel *infopix, 
                                struct infoPixel *infopixB)
{
  int iR, iG, iB;
  struct RGBpoint *point, *pointnext;

  for (iR=0; iR < 256; iR++) {
    for (iG=0; iG < 256; iG++) {
      for (iB=0; iB < 256; iB++) {
        if (check[iR][iG][iB].first) {
          for (point=check[iR][iG][iB].first; point; point=pointnext) {
            pointnext=point->next;
            delete point;
          }
        }
      }
      delete[] check[iR][iG];
    }
    delete[] check[iR];
  }
  delete[] check;
  delete[] infopix;
  if (infopixB) delete[] infopixB;
}


/**
* \brief add RGB points to a list of neighbors of a given RGB point in a color 
*        image
*
* Adds to the list all the colors of the image whose RGB coordinates are in the
* range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
*
* @param[in] R0, G0, B0 initial RGB point
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*            that contains the list of pixels whose color is in the range 
*            [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[in] rneigh threshold distance that defines the neighborhood around the
*            initial RGB point
* @param[in/out] firstneighIn, lastneighIn pointers to the beginning and end
*                of the list
* @param[out] nneigh number of color points in the list (each point appears as 
*             many times as pixels in the image have this color)
* @param[out] nneighdiff number of diferent colors in the list (without taking 
*             into account the number of times it is repeated in the image)
*
*/
void add_neighboursRGB(float R0, float G0, float B0, struct RGBbin ***check, 
                       int iR, int iG, int iB, int rneigh, 
                       struct RGBpoint **firstneighIn, 
                       struct RGBpoint **lastneighIn, 
                       int &nneigh, int &nneighdiff)
{
  struct RGBpoint *firstneigh, *lastneigh;
  struct RGBpoint *point, *pointB;
  unsigned char found;

  firstneigh=*firstneighIn;
  lastneigh=*lastneighIn;

  //add points with coordinates in the range [iR, iR+1), [iG, iG+1), [iB, iB+1)
  for (point=check[iR][iG][iB].first; point; point=point->next) {
    //check that the L1-distance to the initial point is not above 
    //the threshold 
    if ((fabs(point->R-R0) <= rneigh) && (fabs(point->G-G0) <= rneigh) 
         && (fabs(point->B-B0) <= rneigh)) {
      found=0;
      if (!firstneigh) firstneigh=point;
      else {
        //check not already in list
        for (pointB=check[iR][iG][iB].first; 
             (pointB != point) && !found; pointB=pointB->next) {
          found=(pointB->R == point->R) && 
                (pointB->G == point->G) && 
                (pointB->B == point->B);
	  }
        lastneigh->nextneigh=point;
      }
        lastneigh=point;
        point->nextneigh=NULL;
        //increase counter of points
        nneigh++;
        //if not already in list increase counter of different points 
        if (!found) nneighdiff++;
      }
  }

  *firstneighIn=firstneigh;
  *lastneighIn=lastneigh;
}

/**
* \brief create a list of neighbors of a given RGB point (RGB points at 
* L1-distance smaller than a threshold) in a color image
*
* @param[in] R0, G0, B0 initial RGB point
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*                   that contains the list of pixels whose color is in the 
*                   range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[in] rneigh threshold distance that defines the neighborhood around the
*                   initial RGB point
* @param[in] nneighmax if nneighmax > 0, maximum number of different RGB
*                      points in the list, otherwise do not limit the size of
*                      the list
* @param[out] nneigh number of color points in the list (each point appears as 
*                    many times as pixels in the image have this color)
* @param[out] nneighdiff number of diferent colors in the list (without taking 
*                        into account the number of times it is repeated in
*                        the image)
*
* The list is filled by looking for neighbors at L1-distances 
* 0, 1, 2,..., rneigh, until at least nneighmax different colors are 
* found (in the case that nneighmax > 0), 
* or until the maximum distance rneigh is reached 
* (in the case that nneighmax <= 0)
*
* 
* \return pointer to the beginning of the list
*/
struct RGBpoint *find_neighboursRGB(float R0, float G0, float B0, 
                                    struct RGBbin ***check, 
                                    int rneigh, int nneighmax, 
                                    int &nneigh, int &nneighdiff)
{
  struct RGBpoint *firstneigh, *lastneigh;
  int iR, iG, iB;

  //initialize counters and pointers
  nneigh=nneighdiff=0;
  firstneigh=NULL;
  lastneigh=NULL;


  int iR1, iR2, iG1, iG2, iB1, iB2, rn;
  rn=0;
  iR1=iR2=(int) R0;
  iG1=iG2=(int) G0;
  iB1=iB2=(int) B0;
  
  do {
    //check for neighbors at distance rn
    for (iB=iB1; iB <= iB2; iB++) {
      if ((iB >= 0) && (iB <= 255)) {
        if ((iB == iB1) || (iB == iB2)) {
          for (iR=iR1; iR <= iR2; iR++) {
            if ((iR >= 0) && (iR <= 255)) {
              for (iG=iG1; iG <= iG2; iG++) {
                if ((iG >= 0) && (iG <= 255)) {
                  add_neighboursRGB(R0, G0, B0, check, iR, iG, iB, rneigh, 
                                    &firstneigh, &lastneigh, nneigh,
                                    nneighdiff); 
                }
              }
            }
          }
        } else {

          iR=iR1;
          if ((iR >= 0) && (iR <= 255)) {
            for (iG=iG1; iG <= iG2; iG++) {
              if ((iG >= 0) && (iG <= 255)) {
                add_neighboursRGB(R0, G0, B0, check, iR, iG, iB, rneigh, 
                                  &firstneigh, &lastneigh, nneigh,
                                  nneighdiff); 
              }
            }
          }
          iR=iR2;
          if ((iR >= 0) && (iR <= 255)) {
            for (iG=iG1; iG <= iG2; iG++) {
              if ((iG >= 0) && (iG <= 255)) {
                add_neighboursRGB(R0, G0, B0, check, iR, iG, iB, rneigh, 
                                  &firstneigh, &lastneigh, nneigh, nneighdiff);
              }
            }
          }
          iG=iG1;
          if ((iG >= 0) && (iG <= 255)) {
            for (iR=iR1+1; iR <= iR2-1; iR++) {
              if ((iR >= 0) && (iR <= 255)) {
                add_neighboursRGB(R0, G0, B0, check, iR, iG, iB, rneigh, 
                                  &firstneigh, &lastneigh, nneigh,
                                  nneighdiff); 
              }
            }
          }
          iG=iG2;
          if ((iG >= 0) && (iG <= 255)) {
            for (iR=iR1+1; iR <= iR2-1; iR++) {
              if ((iR >= 0) && (iR <= 255)) {
                add_neighboursRGB(R0, G0, B0, check, iR, iG, iB, rneigh, 
                                  &firstneigh, &lastneigh, nneigh,
                                  nneighdiff); 
              }
            }
          }
        }
      }
    }

    //increase rn
    iR1--;
    if (iR1 < 0) iR1=-1;
    iR2++;
    if (iR2 > 255) iR2=256;
    iG1--;
    if (iG1 < 0) iG1=-1;
    iG2++;
    if (iG2 > 255) iG2=256;
    iB1--;
    if (iB1 < 0) iB1=-1;
    iB2++;
    if (iB2 > 255) iB2=256;
    rn++;

  } while ((rn <= rneigh) && ((nneighmax <= 0) || (nneighdiff < nneighmax)));


  return firstneigh;
}


//option=1 -> project 1D
//option=2 -> project 2D
/**
* \brief project RGB point onto a line or a plane
*
* @param[in] R0, G0, B0 coordinates of the initial RGB point
* @param[out] newR, newG, newB coordinates of the projected point
* @param[in] Rcm, Gcm, Bcm point coordinates of a point contained in the line 
*                          or plane
* @param[in] vx, vy, vy coordinates of the unitary direction vector of the
*                       line, or the orthonormal vector to the plane
* @param[in] option option=1 project to line, otherwise project to plane
*
*/
void project_point(float R0, float G0, float B0, 
                   float &newR, float &newG, float &newB, 
                   float Rcm, float Gcm, float Bcm, 
                   float vx, float vy, float vz, 
                   unsigned char option)
{
  float vv[3];
  vv[0]=vx; vv[1]=vy; vv[2]=vz;
  if (option == 1) {
    float dp[3], lambda;
    //project to line 
    dp[0]=R0-Rcm;
    dp[1]=G0-Gcm;
    dp[2]=B0-Bcm;
    lambda=scalar_product(dp, vv);
    newR=Rcm+lambda*vv[0];
    newG=Gcm+lambda*vv[1];
    newB=Bcm+lambda*vv[2];
  } else {
    //project point to plane
    float A, B, C, D, lambda;
    A=vv[0];
    B=vv[1];
    C=vv[2];
    D=-(A*Rcm+B*Gcm+C*Bcm);
    lambda=-(D+R0*A+G0*B+B0*C);
    newR=R0+lambda*A;
    newG=G0+lambda*B;
    newB=B0+lambda*C;
  }
}

//option=1 -> project to line 
//option=2 -> project to plane 
/**
* \brief get parameters of a local line or a plane in the neighborhood of a 
* given RGB point (use PCA) and project the point onto the line (plane)
*
* The spatial distribution of the RGB points in the neighborhood of the 
* considered point (L1-distance below a given threshold), including the point
* itself, is analyzed using Principal Components Analysis. For this analysis, 
* the frequency of the color point (i.e. the number of times it appears in the 
* image) is taken into account
* 
* The output line (or plane) is defined by the first (or two first) principal 
* components and passes through the center of mass of the considered RGB points
*
* @see compute_pca3
*
* @param[in] R0, G0, B0 initial RGB point
* @param[in] firstneigh pointer to the list of neighbors of the input point
* @param[out] infoproj data structure containing the parameters of the line or 
*                      plane and the projection of the initial point
* @param[in] if option=2 compute plane, otherwise compute line
*
* \return 1 if projection to a line, 2 if projection to a plane
*/
unsigned char get_projectionRGB(float R0, float G0, float B0, 
                                struct RGBpoint *firstneigh, 
                                struct infoProjection *infoproj, 
                                unsigned char option)
{
  int nneigh;
  double Rcm, Gcm, Bcm;
  float newR, newG, newB;
  float uu[3], vv[3], ww[3];
  double mean[3], C[3][3], VecPs[3][3], ValPs[3];
  unsigned char projection_type;

  //compute center of mass and covariance matrix of the set of points
  for (int i=0; i < 3; i++) mean[i]=0;
  for (int i=0; i < 3; i++)
    for (int j=0; j < 3; j++) C[i][j]=0;
  nneigh=0;
  for (struct RGBpoint *point=firstneigh; point; point=point->nextneigh) {
    mean[0]+=(double) point->R;
    mean[1]+=(double) point->G;
    mean[2]+=(double) point->B;
    C[0][0]+=((double) point->R)*((double) point->R);
    C[0][1]+=((double) point->R)*((double) point->G);
    C[0][2]+=((double) point->R)*((double) point->B);
    C[1][1]+=((double) point->G)*((double) point->G);
    C[1][2]+=((double) point->G)*((double) point->B);
    C[2][2]+=((double) point->B)*((double) point->B);
    nneigh++;
  }
  for (int i=0; i < 3; i++) mean[i]/=((double) nneigh);
  for (int i=0; i < 3; i++)
    for (int j=i; j < 3; j++) 
      C[i][j]=(C[i][j]/(double) nneigh)-mean[i]*mean[j];
  for (int i=0; i < 3; i++)
    for (int j=0; j < i; j++) C[i][j]=C[j][i];


  //PCA of the set of RGB points
  compute_pca3(C, ValPs, VecPs);
  
  Rcm=(float) mean[0];
  Gcm=(float) mean[1];
  Bcm=(float) mean[2];
  
  //uu and vv are the vectors associated to two first principal components
  uu[0]=(float) VecPs[0][0];
  uu[1]=(float) VecPs[1][0];
  uu[2]=(float) VecPs[2][0];
  vv[0]=(float) VecPs[0][1];
  vv[1]=(float) VecPs[1][1];
  vv[2]=(float) VecPs[2][1];
  normalize_vector(uu);
  normalize_vector(vv);
  
  projection_type=0;
  if ((option == 2) && (ValPs[1] != 0)) { //to prevent projection to a plane
                                          //when points distributed along a
                                          //line
    //project to plane
    vectorial_product(uu, vv, ww);
    normalize_vector(ww);
    
    infoproj->A=ww[0];
    infoproj->B=ww[1];
    infoproj->C=ww[2];
    infoproj->D=-(infoproj->A*Rcm+infoproj->B*Gcm+infoproj->C*Bcm);
    infoproj->Rcm=Rcm;
    infoproj->Gcm=Gcm;
    infoproj->Bcm=Bcm;
   
    //project point to plane
    project_point(R0, G0, B0, newR, newG, newB, Rcm, Gcm, Bcm, 
                    infoproj->A, infoproj->B, infoproj->C, (unsigned char) 2);
    projection_type=2; //projected 2D
  } else {
    //project to line
    project_point(R0, G0, B0, newR, newG, newB, Rcm, Gcm, Bcm, 
                  uu[0], uu[1], uu[2], (unsigned char) 1);
    infoproj->A=uu[0];
    infoproj->B=uu[1];
    infoproj->C=uu[2];
    infoproj->Rcm=Rcm;
    infoproj->Gcm=Gcm;
    infoproj->Bcm=Bcm;
    projection_type=1; //projected 1D
  }
 
  infoproj->Rp=newR;
  infoproj->Gp=newG;
  infoproj->Bp=newB;
    
  return projection_type;
}

/**
* \brief find the projection of a given RGB point, from a color image, onto a
* locally defined line or plane
* 
* The list of neighbors (RGB points at L1-distance below some threshold) of 
* the considered point is first computed. Each RGB point appears in the list as
* many times as it is found in the image (=number of pixels with the same RGB 
* value), however the number of different RGB points is also computed. If the 
* number of different RGB points is large enough then Principal Components 
* Analysis is used to compute the line or plane
* that better fits the spatial distribution of RGB points.
* Finally the original point is projected onto this line or plane.
* 
* @param[in] R0, G0, B0 initial RGB point
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*                   that contains the list of pixels whose color is in the
*                   range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[out] infoproj data structure containing the parameters of the line or 
*                      plane and the projection of the initial point
* @param[in] rneigh threshold distance that defines the neighborhood of the 
*                   initial RGB point
* @param[in] nneighmin minimum number of different RGB points needed to
*                      perform PCA and compute projection
* @param[in] option if option=2 compute plane, otherwise compute line
*
* @see get_projectionRGB
*
* \return 0 if the point could not be projected, otherwise return 1
* (projection onto line) o 2 (projection onto plane)
*/
unsigned char find_projectionRGB(float R0, float G0, float B0, 
                                 struct RGBbin ***check, 
                                 struct infoProjection *infoproj, 
                                 int rneigh, int nneighmin, 
                                 unsigned char option)
{
  int nneigh, nneighdiff;
  struct RGBpoint *firstneigh;
  unsigned char validprojection;

  //get list of neighbors
  nneigh=0;
  firstneigh=find_neighboursRGB(R0, G0, B0, check, rneigh, -1, nneigh,
                                nneighdiff);
  //check if the number of different colors is large enogh
  if (nneighdiff >= nneighmin) {
    //project point to line or plane
    validprojection=get_projectionRGB(R0, G0, B0, firstneigh, infoproj,
                                      option);
  } else validprojection=0;
  
  return validprojection;
}



/**
* \brief update RGB values from the projection of initial RGB points onto local
* lines or planes
*
* @param[in] R0, G0, B0 initial RGB point
* @param[in] validprojection label indicating whether the point is projected 
*                            to a line (1), a plane (2) or is not projected 
*                            at all (0)
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*                   that contains the list of pixels whose color is in the
*                   range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[in] infopix information about the current colors of the pixels  
* @param[out] infopixB information about the updated colors of the pixels  
* @param[out] infoproj data structure containing the parameters of the line or 
*                      plane and the projection of the initial point
* @param[in] rsim speed up parameter, all the points in a neighborhood of
*                 radius rsim of a the initial color point are projected to
*                 the same line/plane
*
*
*/
void process_RGBpoint(float R0, float G0, float B0, 
                      unsigned char validprojection, 
                      struct RGBbin ***check, struct infoPixel *infopix, 
                      struct infoPixel *infopixB, 
                      struct infoProjection *infoproj, 
                      int rsim)
{
  int iR, iG, iB, iRmin, iRmax, iGmin , iGmax, iBmin, iBmax;
  struct RGBpoint *point;

  //neighborhood around initial point
  iRmin=(int) (R0-rsim);
  iRmax=(int) (R0+rsim);
  iGmin=(int) (G0-rsim);
  iGmax=(int) (G0+rsim);
  iBmin=(int) (B0-rsim);
  iBmax=(int) (B0+rsim);

  //check that the neighborhood is inside the color cube
  if (iRmin < 0) iRmin=0;
  if (iRmin > 255) iRmin=255;
  if (iRmax < 0) iRmax=0;
  if (iRmax > 255) iRmax=255;
  if (iGmin < 0) iGmin=0;
  if (iGmin > 255) iGmin=255;
  if (iGmax < 0) iGmax=0;
  if (iGmax > 255) iGmax=255;
  if (iBmin < 0) iBmin=0;
  if (iBmin > 255) iBmin=255;
  if (iBmax < 0) iBmax=0;
  if (iBmax > 255) iBmax=255;

  //process all the points in the neighborhood
  for (iR=iRmin; iR <= iRmax; iR++)
    for (iG=iGmin; iG <= iGmax; iG++)
      for (iB=iBmin; iB <= iBmax; iB++) {

        for (point=check[iR][iG][iB].first; point; point=point->next) {
          if ((fabs(point->R-R0) <= rsim) && (fabs(point->G-G0) <= rsim) 
              && (fabs(point->B-B0) <= rsim)) {
            infopix[point->idpix].processed=1;
            if (validprojection > 0) {
              //if projection (either to line or plane) is possible, 
              //then update color information 
              project_point(point->R, point->G, point->B, 
                            infopixB[point->idpix].R, 
                            infopixB[point->idpix].G, 
                            infopixB[point->idpix].B, 
                            infoproj->Rcm, infoproj->Gcm, infoproj->Bcm, 
                            infoproj->A, infoproj->B, infoproj->C, 
                            validprojection);
              infopixB[point->idpix].AA=infoproj->A;
              infopixB[point->idpix].BB=infoproj->B;
              infopixB[point->idpix].CC=infoproj->C;
              infopixB[point->idpix].DD=infoproj->D;
              infopixB[point->idpix].Rcm=infoproj->Rcm;
              infopixB[point->idpix].Gcm=infoproj->Gcm;
              infopixB[point->idpix].Bcm=infoproj->Bcm;
              infopixB[point->idpix].projected=validprojection;
            } else {
              //if projection is not possible keep the old RGB value
              infopixB[point->idpix].R=R0;
              infopixB[point->idpix].G=G0;
              infopixB[point->idpix].B=B0;
              infopixB[point->idpix].AA=0;
              infopixB[point->idpix].BB=0;
              infopixB[point->idpix].CC=0;
              infopixB[point->idpix].DD=0;
              infopixB[point->idpix].Rcm=0;
              infopixB[point->idpix].Gcm=0;
              infopixB[point->idpix].Bcm=0;
              infopixB[point->idpix].projected=0;
            }
          }
        } //for point

      }

}

/**
* \brief reset data structures for a new iteration of the filtering algorithm 
*
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*                   that contains the list of pixels whose color is in the 
*                   range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
*
*/
void reset_color_info(struct RGBbin ***check)
{
  int iR, iG, iB;
  struct RGBpoint *point, *pointnext;

  for (iR=0; iR < 256; iR++)
    for (iG=0; iG < 256; iG++)
      for (iB=0; iB < 256; iB++) {
        if (check[iR][iG][iB].first) {
          for (point=check[iR][iG][iB].first; point; point=pointnext) {
            pointnext=point->next;
            delete point;
          }
          check[iR][iG][iB].first=check[iR][iG][iB].last=NULL;
        }
      }
}

/**
* \brief compute average difference between RGB values in current and previous 
* iterations of the algorithm
*
* @param[in] infopix, infoPrev information about the colors of the pixel 
*                              in the current and previous iterations of the
*                              algorithm 
* @param[in] w, h dimensions of the input image (w*h = number of pixels)
* @param[out] dprev average difference between current and previous iterations 
*                   of the algorithm 
* @params[in] repeatedRGB array of labels for the image pixels. 
*                         Only one instance of each color has a label
*                         different from zero
*
* This function is used in the main loop to decide whether further 
* iterations are needed (if dprev is above a certain threshold)
*
*/
void compute_errors_it(struct infoPixel *infopix, 
                       struct infoPixel *infopixPrev, 
                       int w, int h, float &dprev, int *repeatedRGB)
{
  float dp;

  dprev=0;
  int ndiff=0;
  for (int n=0; n < w*h; n++) {
    if (repeatedRGB[n] > 0) { //compute differences in the color cloud
      //count number of different colors 
      ndiff++;
      //difference between the value of pixel n in the current 
      //and previous iterations
      dp=sqrt((infopixPrev[n].R-infopix[n].R)*(infopixPrev[n].R-infopix[n].R)+
              (infopixPrev[n].G-infopix[n].G)*(infopixPrev[n].G-infopix[n].G)+
              (infopixPrev[n].B-infopix[n].B)*(infopixPrev[n].B-infopix[n].B));
      dprev+=dp;
    }
  }
  //This line of code has been modified on July 5, 2011 
  //to fix a bug that occurs on images whose colors are 
  //sparsely distributed in RGB space.
  //Old line:
  //dprev/=(float) ndiff;
  //New line:
  if (ndiff > 0) dprev/=(float) ndiff;
}

/**
* \brief update color information for a new iteration of the filtering algorithm 
*
* @param[in] check  check[iR][iG][iB] is a pointer to a RGBbin data structure 
*                   that contains the list of pixels whose color is in the
*                   range [iR, iR+1), [iG, iG+1), [iB, iB+1) 
* @param[in] infopix information about the colors of the pixel 
* @param[in] w, h dimensions of the input image (w*h = number of pixels)
*
*/
void update_color_info(struct RGBbin ***check, struct infoPixel *infopix, 
                       int w, int h)
{
  int n, nRGB, iR, iG, iB;
  struct RGBpoint *point, *pointB;
  unsigned char isrepeated;

  nRGB=0;
  for (n=0; n < w*h; n++) {
    if (!infopix[n].ignore) {
      infopix[n].processed=0;
	//update lists of color points
      point=new struct RGBpoint;
      point->R=infopix[n].R;
      point->G=infopix[n].G;
      point->B=infopix[n].B;
      point->idpix=n;
      point->next=NULL;
      //index of the color point in data structure
      iR=(int) infopix[n].R;
      iG=(int) infopix[n].G;
      iB=(int) infopix[n].B;
      if (iR < 0) iR=0;
      if (iR > 255) iR=255;
      if (iG < 0) iG=0;
      if (iG > 255) iG=255;
      if (iB < 0) iB=0;
      if (iB > 255) iB=255;
      if (check[iR][iG][iB].first == NULL) nRGB++;
      else {
        isrepeated=0;
        for (pointB=check[iR][iG][iB].first; pointB && !isrepeated; 
             pointB=pointB->next) {
          isrepeated=(pointB->R == point->R) && (pointB->G == point->G) 
                     && (pointB->B == point->B);
	    }
        //count number of different colors
        if (!isrepeated) nRGB++;
      }
      if (check[iR][iG][iB].first == NULL) check[iR][iG][iB].first=point;
      else check[iR][iG][iB].last->next=point;
      check[iR][iG][iB].last=point;
    }
  }
  printf("   %i colors\n", nRGB);
}

/**
* \brief compute output image from updated RGB values 
*
* @param[in] outR, outG, outB color channels of the output image
* @param[in] w, h dimensions of the output image
* @param[in] infopix information about the colors of the pixel 
*
*/
void get_output_image(unsigned char *outR, unsigned char *outG, 
                      unsigned char *outB, int w, int h, 
                      struct infoPixel *infopix)
{
  int iR, iG, iB, n, nnoprojected=0, nignored=0;

  for (n=0; n < w*h; n++) {
    //round to the closest integer values in the range [0, 255]
    iR=(int) (infopix[n].R + 0.5f);
    iG=(int) (infopix[n].G + 0.5f);
    iB=(int) (infopix[n].B + 0.5f);
    if (iR < 0) iR=0;
    if (iR > 255) iR=255;
    if (iG < 0) iG=0;
    if (iG > 255) iG=255;
    if (iB < 0) iB=0;
    if (iB > 255) iB=255;
    if (infopix[n].ignore) nignored++;
    else {
      if (!infopix[n].projected) nnoprojected++;
    }
    //optional: assign (0, 0, 0) to non-projected pixels
    //if (!infopix[n].projected) { iR=iG=iB=0; }
    //
    outR[n]=iR;
    outG[n]=iG;
    outB[n]=iB;
  }
  printf("%i (%2.2f) ignored\n", 
         nignored, 100.0f* (float) nignored/(float) (w*h));
  printf("%i (%2.2f) non-projected\n", 
         nnoprojected, 100.0f* (float) nnoprojected/(float) (w*h));
}


//option=1 -> project to line 
//option=2 -> project to plane 
/**
* \brief Main function implementing the color filtering algorithm (LLP algorithm)
*
* Algorithm description (LLP algorithm)
*
* Image colors are considered as points in a 3D space, namely the color cube.
* The algorithm replaces each RGB color in the cube by its projection onto a 
* line (or plane) locally defined. 
* This line (or plane) is the one that better fits the spatial distribution of 
* the colors in a neighborhood of the considered RGB point, and it is computed 
* by Principal Components Analysis (PCA).
* Only points with a large enough number of neighboring colors are projected. 
* The process is iterated until the average error between corresponding colors 
* in consecutive iterations is below a threshold.
*
* @param[in] R, G, B pointers to red, green and blue channels of the input
*                    image
* @param[out] outR, outG, outB  pointers to red, green and blue channels of
*                               the output image
* @param[in]  w, h   width and height of the input (and output) images
* @param[in]  param  data structure with the parameters of the algorithm:
* @param  param.rneigh radius of local neighborhood for computation of lines 
*                      or planes
* @param  param.nneighmin minimum number of neighbors (number of different 
*                         colors) needed to compute lines or planes
* @param  param.rsim speed up factor, the same line or plane computed for a 
*                    given RGB point is used for its neighbors at L1-distance 
*                    @f$ \leq @f$ rsim
* @param  param.eitmax stop criterion of the algorithm. 
*                      The algorithm is iterated until the average 
*                      difference of corresponding colors in two consecutive
*                      iterations is below eitmax 
* @param  param.option 1 project to line, 2 project to plane
* @param  param.ignore000 if TRUE ignore (0,0,0) color points (eventually 
*                         belonging to pixels discarded in a pre-processing
*                         step)
*
*
* \return number of iterarations of the algorithm
*
*/
int colorfiltering(unsigned char *R, unsigned char *G, unsigned char *B, 
                   unsigned char *outR, unsigned char *outG, 
                   unsigned char *outB,
                   int w, int h, struct paramColorFilter *param)
{
  //input parameters
  int rneigh, rsim, nneighmin;
  float eitmax;
  unsigned char option, ignore000;

  //get parameters from data structure
  rneigh=param->rneigh;
  nneighmin=param->nneighmin;
  rsim=param->rsim;
  eitmax=param->eitmax;
  option=param->option;
  ignore000=param->ignore000;

  //auxiliary data structures
  struct RGBbin ***check;
  struct infoPixel *infopix, *infopixB, *infopixAux;
  struct infoProjection infoproj;
  int n, it;
  float R0, G0, B0;
  unsigned char validprojection;
  float eit;
  int *repeatedRGB=NULL;

  //initialization of data structures
  initialize_colorfiltering_data(R, G, B, w, h, &check, &infopix, &infopixB, 
                                 &repeatedRGB, ignore000);

  //start filtering
  unsigned char stop;
  it=0;
  do {
    printf("it=%i\n", it);
    for (n=0; n < w*h; n++) {
      if (infopix[n].ignore) {
        infopixB[n].ignore=1; //do not process pixels tagged as "ignored"
        infopixB[n].R=infopix[n].R;
        infopixB[n].G=infopix[n].G;
        infopixB[n].B=infopix[n].B;
      } else {
        infopixB[n].ignore=0;
        if (!infopix[n].processed) { //do not process already processed colors
          R0=infopix[n].R;
          G0=infopix[n].G;
          B0=infopix[n].B;

          //find projection
          validprojection=find_projectionRGB(R0, G0, B0, check, &infoproj, 
                                             rneigh, nneighmin, option);
 
          //mark as processed
          process_RGBpoint(R0, G0, B0, validprojection, check, infopix, 
                           infopixB, &infoproj, rsim);

        } //if !infopix[n].processed
      } //if !infopix[n].ignore
    } //for n

    //compute errors
    compute_errors_it(infopixB, infopix, w, h, eit, repeatedRGB);
    printf("   eit=%2.4f\n", eit);

    //update info pix
    infopixAux=infopix;
    infopix=infopixB;
    infopixB=infopixAux;

    //reset color info
    reset_color_info(check);

    //update color info
    update_color_info(check, infopix, w, h);

    it++;
    stop=(eit < eitmax);
  } while (!stop);

  //output image
  get_output_image(outR, outG, outB, w, h, infopix);

  //clean up  
  delete_colorfiltering_data(check, infopix, infopixB);
  delete[] repeatedRGB;

  return it;
}
					 
/**
* \brief Organize input parameters in a data structure
*
* @param[in] rneigh distance defining the neighborhood of a color point 
*                   (neighbors are RGB points at L1-distance below rneigh)
* @param[in] rsim distance defining a small neighborhood around each color 
*                 point, it is used to speed up the algorithm. It must be 
*                 much smaller than rneigh, typically rsim=rneigh/3 
* @param[in] nneighmin minimum number of neighbors needed to compute a local 
*                      line or plane in the vicinity of a color point 
* @param[in] eitmax stop criterion of the algorithm. The algorithm is iterated 
*                   until the average difference of corresponding colors in 
*                   two consecutive iterations is below eitmax 
* @param[in] option project points to lines (option=1) or planes (option=2) 
* @param[in] ignore000 if set to 1, do not take into account pixels with
*                      (0,0,0) RGB value (eventually discarded for computation
*                      in a preprocessing step) 
*
*
* \return Data structure containing the input parameters
*/
struct paramColorFilter *new_colorfiltering_parameters(int rneigh, int rsim, 
                                                       int nneighmin, 
                                                       float eitmax, 
                                                       unsigned char option, 
                                                       unsigned char ignore000)
{
  struct paramColorFilter *param=new struct paramColorFilter;
  param->rneigh=rneigh;
  param->rsim=rsim;
  param->nneighmin=nneighmin;
  param->eitmax=eitmax;
  param->option=option;
  param->ignore000=ignore000;

  return param;
}

/**
* \brief Delete input parameters data structure
*
*
*/
void delete_colorfiltering_parameters(struct paramColorFilter *param)
{
  delete param;
}


/**
* \brief Compute the density of the colors of an image
*
* The density of a color C (in RGB space) is defined as 
*
* @f[
* \delta(C)=\sum_{C' \in {\cal S}} e^{-d(C, C')^2/h^2} 
* @f]
*
* where @f$ {\cal S} @f$ is the set of neighbors of C 
* @f$ d(C, C') @f$ is the Euclidean distance between colors C and C' 
* in RGB space and @f$ h @f$ is a weighting parameter
*
* The set of neighbors of a color C is composed of all the colors at 
* L1-distance below some threshold @f$ r @f$.
* Each color appears in the set of neighbors as many times as it is present in 
* the image (i.e. the frequency of the color is taken into account in the 
* computations).
* Tipically the weighting parameter is @f$ h = r/\sqrt{2} @f$ so the weight of 
* the furthest neighbors is @f$ e^{-2}=0.135 @f$
*
* @param[in] R,G,B color channels of the image
* @param[in] w,h dimensions of the image
* @param[in] rdst radius of the neighbor around each color point
* @param[in] hdst weighting parameter
* @param[out] maxdst maximum density value
*
* \return density of each RGB point of the image
*/
float ***RGBdensities(unsigned char *R, unsigned char *G, unsigned char *B, 
                      int w, int h, int rdst, float hdst, float &maxdst)
{
  struct RGBbin ***check;
  struct infoPixel *infopix;
  struct RGBpoint *firstneigh, *neigh, *point;
  int iR, iG, iB, nneigh, nneighdiff;
  float ***dsts=NULL;
  float dR, dG, dB, d2, h2, dst;

  //initialization
  initialize_colorfiltering_data(R, G, B, w, h, &check, &infopix, 
                                 NULL, NULL, 1); //ignore (0, 0, 0) pixels
  dsts=new float**[256];
  for (iR=0; iR < 256; iR++) {
    dsts[iR]=new float*[256];
    for (iG=0; iG < 256; iG++) {
      dsts[iR][iG]=new float[256];
      memset(dsts[iR][iG], 0, 256*sizeof(float));
    }
  }  

  //compute densities
  h2=hdst*hdst;
  maxdst=0;
  for (int n=0; n < w*h; n++) {
    if (!infopix[n].ignore && !infopix[n].processed) {
      //get list of neighboring colors
      firstneigh=find_neighboursRGB(infopix[n].R, infopix[n].G, infopix[n].B, 
                                    check, rdst, -1, nneigh, nneighdiff);
      //compute density
      dst=0;
      for (neigh=firstneigh; neigh; neigh=neigh->nextneigh) {
        dR=infopix[n].R-neigh->R;
        dG=infopix[n].G-neigh->G;
        dB=infopix[n].B-neigh->B;
        d2=dR*dR+dG*dG+dB*dB;
        dst+=exp(-d2/h2);
      }
      //store density values
      iR=(int) infopix[n].R;
      iG=(int) infopix[n].G;
      iB=(int) infopix[n].B;
      dsts[iR][iG][iB]=dst;
      //update maximum density value
      if (dst > maxdst) maxdst=dst;
        for (point=check[iR][iG][iB].first; point; point=point->next) 
          infopix[point->idpix].processed=1;  
      }
  }

  //clean up  
  delete_colorfiltering_data(check, infopix, NULL);

  return dsts;
}

/**
* \brief Delete data structure containing density values
*
*/
void delete_RGBdensities(float ***dsts)
{
  int iR, iG;

  for (iR=0; iR < 256; iR++) {
    for (iG=0; iG < 256; iG++) {
      delete[] dsts[iR][iG];
    }
    delete[] dsts[iR];
  }
  delete[] dsts;
}


/**
* \brief Filter out "isolated" colors in RGB space, that is, colors with a
* small number of neighbors 
*
* The result is an image identical to the original, except at the pixels with 
* "isolated" colors, whose value has been replace by (0, 0, 0).
* The number of neighbors of a pixel is the number of color points at 
* L1-distance below some threshold.
* Each color appears in the list of neighbors as many times as it is 
* present in the image (i.e. the frequency of the color is taken into 
* account in the computations).
*
* @param[in] R, G, B color channels of the input image
* @param[out] outR, outG, outB color channels of the output image
* @param[in] w, h dimensions of input and output images
* @param[in] r radius of the neighborhood around each color point
* @param[in] nmin threshold for the minimum number of neighboring color points 
*                 needed to decide that a color is not "isolated"
* @param[out] nRGB number of colors in input image
* @param[out] nisolatedRGB number of "isolated" colors 
* @param[out] nisolatedpixels number of pixel with "isolated" color
*
*/
void removeisolatedRGB(unsigned char *R, unsigned char *G, unsigned char *B, 
                       unsigned char *outR, unsigned char *outG, 
                       unsigned char *outB,
                       int w, int h, int r, int nmin, int &nRGB, 
                       int &nisolatedRGB, 
                       int &nisolatedpixels)
{
  struct RGBbin ***check;
  struct infoPixel *infopix;
  struct RGBpoint *firstneigh, *point;
  int iR, iG, iB, nneigh, nneighdiff;

  nRGB=0;
  nisolatedRGB=0;
  nisolatedpixels=0;

  memcpy(outR, R, w*h);
  memcpy(outG, G, w*h);
  memcpy(outB, B, w*h);

  //initialization
  nRGB=initialize_colorfiltering_data(R, G, B, w, h, &check, &infopix, 
                                      NULL, NULL, 1); //ignore (0, 0, 0) pixels

  //compute number of neighbours
  for (int n=0; n < w*h; n++) {
    if (!infopix[n].ignore && !infopix[n].processed) {
      //get list of neighbors
      firstneigh=find_neighboursRGB(infopix[n].R, infopix[n].G, infopix[n].B, 
                                    check, r, nmin, nneigh, nneighdiff);
      iR=(int) infopix[n].R;
      iG=(int) infopix[n].G;
      iB=(int) infopix[n].B;
      //count number of isolated points
      if (nneighdiff < nmin) nisolatedRGB++;
      for (point=check[iR][iG][iB].first; point; point=point->next) {
        infopix[point->idpix].processed=1;  
        if (nneighdiff < nmin) { //set to (0, 0, 0) pixels with isolated colors
          outR[point->idpix]=0;
          outG[point->idpix]=0;
          outB[point->idpix]=0;
          nisolatedpixels++;
        }
      }
    }
  }

  //clean up  
  delete_colorfiltering_data(check, infopix, NULL);

  printf("%i isolated colors (%2.2f%% of all colors) \n", 
         nisolatedRGB, (100.0f*(float) nisolatedRGB)/((float) nRGB));
  printf("%i pixels with isolated colors (%2.2f%% of all pixels) \n", 
         nisolatedpixels, (100.0f*(float) nisolatedpixels)/((float) (w*h)));

}


